# Get a table from the template
getTable <- function(configTable, excelTemplate){
  sheet <- configTable$value$sheet
  row <- configTable$value$row
  col <- configTable$value$col
  name <- configTable$columns$Name
  isCorrelation <- configTable$indicator$isCorrelation
  keyword <- configTable$keyword
  
  A <- getValue(excelTemplate, sheet, objectCategory = "sheet")
  
  if(keyword == "correlationMarket"){
    RF <- A[row - 1, seq(from = col[2], to = ncol(A))]
    if(anyNA(RF)){
      RF <- RF[seq_len(min(which(is.na(RF)))-1)]
    }
    n <- length(RF) + 1
    col <- col[seq_len(n)]
    name <- name[seq_len(n)]
  }
  
  if(configTable$indicator$isValue){
    a <- ifelse(row > nrow(A) || col > ncol(A), NA_character_, A[row, col])
    A <- data.table(a)
    names(A) <- name
    A[, rowNumber := 1]
    
  }else{
    missingCols <- name[col > ncol(A)]
    if(isCorrelation){
      addError(keyword, paste0("The column '", missingCols, "' is missing in the correlation matrix") , columns = missingCols)
    }
    addError(keyword, "The column should not be empty", columns = missingCols)
    A <- data.table(A[, col, drop = FALSE])
    # A <- data.table(A)[, col, with = FALSE]
    
    names(A) <- name
    A[, rowNumber := seq_len(.N)]
    A <- A[rowNumber >= unique(row)]
  }
  return(A)
}


# Get the structure of the Excel file. It specified the coordinated of each table and the column types.
getStructure <- function(excelStructure, colRange){
  
  column <- list(
    AutoComplete = excelStructure$name[excelStructure$characteristic == "AutoComplete"],
    Skip = excelStructure$name[excelStructure$dataSelection == "Skip"],
    FilterValue = excelStructure$filterValue[excelStructure$filterValue != ""],
    FilterName = excelStructure$name[excelStructure$filterValue != ""],
    Split = excelStructure$name[excelStructure$characteristic == "Split"],
    NamedVectorValue = excelStructure$name[excelStructure$characteristic == "Named vector"],
    NamedVectorName = excelStructure$name[excelStructure$characteristic == "Primary key"]
  )
  
  columns <- list(
    ValueRequired = excelStructure$name[excelStructure$isValueRequired == TRUE],
    ReplacementZero = excelStructure$name[excelStructure$isReplacementZero == TRUE],
    PrimaryKey = excelStructure$name[excelStructure$characteristic == "Primary key"],
    Name = excelStructure$name,
    Transpose = excelStructure$name[excelStructure$characteristic == "Transpose"]
  )
  
  value <- list(
    Filter = excelStructure$filterValue[!is.na(excelStructure$filterValue)],
    sheet = excelStructure$sheet[1],
    description = excelStructure$description[1],
    row = excelStructure$row[1],
    col = excelStructure$col
  )
  
  indicator <- list(
    isEmptyAllowed = excelStructure$isEmptyAllowed[1],
    isDropEmptyRows = excelStructure$isDropEmptyRows[1],
    isValue = excelStructure$isValue[1],
    isCorrelation = excelStructure$isCorrelation[1]
  )
  
  keyword <- excelStructure$keyword[1]
  
  range <- colRange[[keyword]]
  if(is.null(range)){
    ranges <- list()
  }else{
    ranges <- lapply(seq_len(nrow(range)), function(i){
      as.list(range[i])
    })
  }
  
  colTypes <- lapply(1:nrow(excelStructure), function(i){list(name = excelStructure$name[i], type = excelStructure$type[i])})
  
  x <- list(keyword = keyword, column = column, columns = columns, value = value, indicator = indicator, ranges = ranges, colTypes = colTypes)
  x <- namedObject(x, "Structure of the Excel file", "Element")
  return(x)
}


# Load an Excel sheet
loadExcelSheet <- function(excelSheet, numberLoad = NULL, path){
  sheet <- excelSheet$sheet
  
  # This parameter is used to avoid loading the whole worksheet in the case the cell in bottom right part of the template is non-empty
  maxCol <- excelSheet$maxCol
  
  # suppressMessages is used to avoid any unwanted message regarding automatic change of column names by tibble
  valExcelTemplate <- unname(suppressMessages(as.matrix(readxl::read_excel(path, sheet, range = readxl::cell_limits(c(1, 1), c(NA, maxCol)), col_names = FALSE, col_types = "text"))))
  
  # Increase the progress bar
  if(!is.null(numberLoad)){
    addProgress(step = 1/numberLoad, message = "Loading") 
  }
  return(valExcelTemplate)
}


loadInitialValue <- function(path, sheet, row, col, replacement){
  tryCatch({
    range <- readxl::cell_limits(c(row, col), c(row, col))
    a <- suppressMessages(suppressWarnings(as.integer(readxl::read_excel(path = path, sheet = sheet, range = range, col_names = FALSE))))
    if(length(a)==1 && !is.null(a) && !is.na(a)){
      a <- floor(a/1e5)
      return(a)
    }
    stop()
  },
  error = function(e){return(replacement)})
}

checkTemplateVersion <- function(templateVersion, sheets){
  messageDetails <- "More details and guidance about the R-Tool are available on FINMA's website."
  
  if(templateVersion == 15910){
    # The R-Tool is compatible with the current version of the SST-Template
  }else if(templateVersion > 15910){
    addError(NA, paste0("A newer version of the R-Tool is required to process this SST-Template. Please use the latest version of the R-Tool provided by FINMA. ", messageDetails))
 
  }else if("list_of_sheets" %in% sheets) {
    addError(NA, paste0("The template provided is a SST-Template for the SST 2020 (or SST 2019). For an official solvency calculation for SST 2020 (or SST 2019), please use the application 'SST-Tool' or the R-package 'sstModel'. ",
                        "Otherwise, please use the SST-Template published together with this R-Tool. ",
                        messageDetails), condition = "list_of_sheets" %in% sheets)
    
  }else{
    addError(NA, paste0("The template provided is not a valid SST-Template. ", messageDetails))
    
  }
}

# Get objects from an Excel file
getObjects <- function(path, variant, numberLoad){
  
  # FINMA only: Use a different configuration file when needed
  externalConfigurationPath <- getOption("sstCalculation.externalConfigurationPath")
  if(!is.null(externalConfigurationPath)){   
    sheets = unique(c(readxl::excel_sheets(path), readxl::excel_sheets(externalConfigurationPath)))
  }else{
    sheets = readxl::excel_sheets(path)
    
    # Check that the correct template version is provided
    templateVersion <- loadInitialValue(path, "Configuration", 2, 26, 0)
    checkTemplateVersion(templateVersion, sheets)
  }
  
  # Determine which sheets are associated to the variant calculation
  sheetsMapping <- getSheetsMapping(sheets = sheets, variant = variant)
  
  # Load the configuration sheet. It contains all informations about the structure of the SST-Template
  excelstructure <- getExcelStructure(path, sheetsMapping)
  
  # Load all relevant worksheets from the SST-Template
  excelTemplate <- getExcelTemplate(path, excelstructure, numberLoad)
  
  # Extracts objects from the SST-Template (correlation matrix, tables, values, etc.)
  excelObjects <- getExcelObjects(excelTemplate, excelstructure)
  
  # Extract data (for the FDS) from the correct variant of the worksheet
  excelObjects$configurationFDS[sheetsMapping, sheetSource := i.newSheet, on = c("sheetSource" = "sheet")]
  excelObjects$configurationFDS[action %in% c("addTextFromExcel", "addNumericFromExcel"), textValue := getTemplateValue(excelTemplate, sheet = sheetSource, row = row2, col = col2), by = "rowNumber"]
  
  # Prepare data export for the FDS in the right FDS variant
  variantSheetName <- ifelse(variant == "", "", paste0("__", variant))
  excelObjects$configurationFDS[, sheet := paste0(sheet, variantSheetName)]
  
  # We sometimes need to use the name of the variant later on for analysis
  excelObjects$variant <- variant
  
  return(excelObjects)
}


# Access to a cell from the SST-Template, which will be  used for the production of the FDS.
getTemplateValue <- function(excelTemplate, sheet, row, col){
  A <- getValue(excelTemplate, sheet, objectCategory = "worksheet")
  if(row > nrow(A) | col > ncol(A)){
    addError(NA, paste0("Can't access to the row ", row, " and column ", col, " from worksheet '", sheet, "'."), type = "Warning")
    return(NA_character_)
  }else{
    return(A[row, col])
  }
}


# Identify the sheets associated to the variant
getSheetsMapping <- function(sheets, variant){
  sheetsMapping <- data.table(rawName = sheets, sheetName = sheets)
  sheetsMapping[, isVariant := paste0("__", variant) == substr(rawName, nchar(rawName) - 3, nchar(rawName))]
  sheetsMapping[isVariant == TRUE , sheetName := substr(rawName, 1, nchar(rawName)-4)]
  sheetsRemove <- sheetsMapping[isVariant == TRUE, sheetName]
  sheetsMapping[!rawName %in% sheetsRemove, .(sheet = sheetName, newSheet = rawName)]
}


# Initialization phase. The configuration from the Excel file is loaded
getExcelStructure <- function(path, sheetsMapping){
  # Structure
  excelStructure <- loadExcelStructure(path)
  excelStructure[sheetsMapping, sheet := i.newSheet, on = "sheet"]
  
  # Load the company name, which will be displayed in the error messages
  companyName <- getCompanyNameRawStructure(path, excelStructure = excelStructure[keyword == "companyName"])
  
  # Load the risk factors to identify the exact size of the correlation matrix
  correlationMarket <- getAdjustedCorrelationMarketStructure(path, correlationMarketStructure = excelStructure[keyword == "correlationMarket"])
  
  # Perform the adjustments
  excelStructure <- rbind(excelStructure[keyword != "correlationMarket"], correlationMarket)
  excelStructure[, company := companyName]
  
  # Save structure in the package environment, which is used to generate meaningful error messsages
  packageEnv$excelStructure <- excelStructure
  
  return(excelStructure)
}


# Get all objects from a template
getExcelObjects <- function(excelTemplate, excelStructure, variant){
  excelStructureSplit <- split(excelStructure, by = "keyword")
  objectsStructure <- lapply(excelStructureSplit, getStructure, colRange = getColRange(excelStructure))
  
  # First load the data as a table
  Tables <- lapplyCatch(objectsStructure, getTable, excelTemplate = excelTemplate)
  
  # Then process the table to the object
  Objects <- lapplyCatch(objectsStructure, checkTable, Tables = Tables)
  
  excelObjects <- namedObject(Objects, "Objects for the SST Model", "Object")
  return(excelObjects)
}


# Load a hard-coded range in the SST-Template with the configuraiton for the R-Tool
loadExcelStructure <- function(path){
  # Used by FINMA only
  externalConfigurationPath <- getOption("sstCalculation.externalConfigurationPath")
  if(!is.null(externalConfigurationPath)){
    path <- externalConfigurationPath
  }
  
  addError(NA, "The Sheet 'Configuration' is missing, please ensure that you are using a valid SST-Template for the SST calculation.", condition = !"Configuration" %in% readxl::excel_sheets(path))
  A <- as.data.table(suppressMessages(readxl::read_excel(path, sheet = "Configuration", range = readxl::cell_limits(c(1, 1), c(NA, 18)), col_names = TRUE, col_types = "text")))
  
  # Cast the columns in the correct type
  for(x in c("row", "col")){
    A[, (x) := as.integer(A[[x]])]
  }
  for(x in c("isValueRequired", "isReplacementZero", "isValue", "isEmptyAllowed", "isDropEmptyRows", "isCorrelation")){
    A[, (x) := as.logical(A[[x]])]
  }
  for(x in c("characteristic", "dataSelection", "filterValue")){
    A[is.na(A[[x]]), (x) := ""]
  }
  return(A[])
}


# Load the whole Excel template
getExcelTemplate <- function(path, excelStructure, numberLoad){
  enableFileCache <- identical(getOption("sstCalculation.enableFileCache"), TRUE)
  externalConfigurationPath <- getOption("sstCalculation.externalConfigurationPath")
  
  excelSheets <- readxl::excel_sheets(path)
  excelSheetsToLoad <- excelStructure[sheet %in% excelSheets, .(maxCol = max(col)), by = "sheet"]
  excelSheetsToLoad <- split(excelSheetsToLoad, by = "sheet")
  
  # Caching
  if(enableFileCache){
    excelTemplatePreLoad <- specialLoadCachedTemplate(path, numberLoad)
    excelTemplate <- lapplyCatch(excelSheetsToLoad, function(x){excelTemplatePreLoad[[x$sheet]]})
  }else{
    excelTemplate <- lapplyCatch(excelSheetsToLoad, loadExcelSheet, path = path, numberLoad = numberLoad * length(excelSheetsToLoad))
  }
  
  # External configuration
  if(!is.null(externalConfigurationPath)){
    excelSheets <- readxl::excel_sheets(externalConfigurationPath)
    excelSheetsToLoad <- excelStructure[sheet %in% excelSheets, .(maxCol = max(col)), by = "sheet"]
    excelSheetsToLoad <- split(excelSheetsToLoad, by = "sheet")
    
    for(x in excelSheetsToLoad){
      excelTemplate[[x$sheet]] <- loadExcelSheet(excelSheet = x, path = externalConfigurationPath, numberLoad = NULL)
      # message("Loaded", x$sheet)
    }
  }
  
  
  excelTemplate <- namedObject(excelTemplate, "SST-Template", "Worksheet")
  
  return(excelTemplate) 
}


# Update the configuration for the correlation matrix to take into account any company-specific adjustment
getAdjustedCorrelationMarketStructure <- function(path, correlationMarketStructure){
  # The first risk factor is the second element
  u <- correlationMarketStructure[2]
  
  # Get the labels from the header row
  RFLabel <- as.character(suppressMessages(readxl::read_excel(path, sheet = u$sheet, range = readxl::cell_limits(c(u$row-1, u$col), c(u$row-1, NA)), col_names = FALSE, col_types = "text")))
  nbRF <- min(which(is.na(c(RFLabel, NA))))-1
  
  correlationMarketStructure <- correlationMarketStructure[1:(nbRF+1)]
  correlationMarketStructure <- correlationMarketStructure[-1, name := RFLabel[1:nbRF]]
  return(correlationMarketStructure)
}


# Get the company name from the Excel file
getCompanyNameRawStructure <- function(path, excelStructure){
  range <- readxl::cell_limits(c(excelStructure$row, excelStructure$col), c(excelStructure$row, excelStructure$col))
  companyName <- paste0(as.character(suppressMessages(readxl::read_excel(path, sheet = excelStructure$sheet, range = range, col_names = FALSE, col_types = "text"))), "")
  return(companyName)
}


# Get a table with one entry per row for checking that the values in a column are in the allowed range
getColRange <- function(excelStructure){
  colRange <- melt(excelStructure, id.vars = c("keyword", "name"), measure.vars = c("Warning", "Error"), variable.name = "type", na.rm = TRUE, variable.factor = FALSE)
  colRange <- colRange[, .(range = unlist(tstrsplit(value, ","))), by = c("keyword", "name", "type")]
  colRange <- split(colRange, by = "keyword")
  return(colRange)
}