# Filter a table to remove all trailing rows.
transformRemoveTrailingRows <- function(A, columns){
  colName <- setdiff(names(A), c(columns, "rowNumber"))
  index <- rowSums(!is.na(A[, colName, with = FALSE])) != 0
  A <- A[seq_len(max(c(0, which(index)))), ]
  return(A)
}

# Filter a table to remove all empty rows.
transformRemoveEmptyRows <- function(A, indicator){
  if(indicator){
    index <- rowSums(!is.na(A[, -"rowNumber"])) != 0
    A <- A[index, ]
  }
  return(A)
}

# Fill down with the same value while the value is not NA
transformAutoComplete <- function(A, columns){
  if(length(columns) == 1){
    v <- A[[columns]]
    for(i in seq_len(length(v)-1)){
      if(is.na(v[i+1])){
        v[i+1] <- v[i]
      }
    }
    A[[columns]] <- v
  }
  return(A)
}

# Filter a table to keep only the rows matching a filter condition, which is specified through a column name and the value it should contain.
transformFilter <- function(A, columns, filter){
  addError(NA, "Internal error: incorrect filtering", condition = !(length(columns) == 1) == (length(filter) == 1))
  if(length(filter) == 1){
    index <- grepl(pattern = filter, x = A[[columns]])
    A <- A[index, ]
  }
  return(A)
}


# Replace NA by 0
transformReplaceNA <- function(A, columns){
  # Warning: it will also modify the object by reference
  # A <- copy(A)
  for(colName in columns){
    addError(NA, "Internal error: Replacement of NA over non-character columns", condition = class(A[[colName]]) != "character")
    index <- is.na(A[[colName]])
    A[index, (colName) := "0"]
  }
  return(A[])
}


# Split the rows of a data.table whenever a ',' is found in the specified column
transformSplitRows <- function(A, columns, keyword = NA){
  if(length(columns) == 1){
    addError(NA, "Internal error: The selected column should be a character column", condition = !is.character(A[[columns]]))
    addErrorCell(keyword, "Invalid input, the cell should not contain the string ',,'", columns = columns, rows = A$rowNumber[grepl(",,",A[[columns]], fixed = TRUE)])
    error <- substr(A[[columns]], 1, 1) == "," | substr(A[[columns]], sapply(A[[columns]], nchar), sapply(A[[columns]], nchar)) == ","
    addErrorCell(keyword, "Invalid input, the cell should not begin or end with ','", columns = columns, rows = A$rowNumber[error])
    
    A <- A[, lapply(.SD, function(x) trimws(unlist(tstrsplit(x, ",", fixed=TRUE)))), by = setdiff(names(A), columns), .SDcols = columns]
  }
  return(A)
}


# Transpose from narrow to wide
transformTranspose <- function(A, columns){
  if(length(columns) == 0){
    return(A)
  }
    
  A <- melt(data = A, measure.vars = columns, value.name = "value", variable.name = "time", variable.factor = FALSE)
  A[, time := as.integer(time)]
  return(A[])
}


# Transform into a correlation matrix
transformCorrelation <- function(A, condition, keyword){
  correlationTolerance <- getParam("correlationTolerance")
  
  if(condition){
    rowNames <- A$name
    rowNumber <- A$rowNumber
    A <-as.matrix(A[, -c("name", "rowNumber")])
    rownames(A) <- rowNames
    
    addError(keyword, paste0("The correlation matrix needs to be a square matrix. The input provided has ", nrow(A), " rows and ", ncol(A), " columns"), columns = "name", condition = nrow(A) != ncol(A))
    
    addMatrixError <- function(U, message){
      u <- which(U > correlationTolerance, arr.ind = TRUE)
      error <- list(row = u[, "row"], col = colnames(A)[u[, "col"]])
      addErrorCell(keyword, message, rows = rowNumber[error$row], columns = error$col)
    }
    
    # Ensure that the correlation matrix is well defined
    addMatrixError(diag(abs(diag(A) - 1)), "Diagonal terms must be equal to '1'")
    addMatrixError(abs(A - t(A)), "Non-symmetric matrix")
    addMatrixError(abs(A) -1, "Correlation larger than 1")
    
    A <- (A + t(A))/2
    diag(A) <- 1
    
    indicator <- !all(eigen(A, symmetric = T, only.values = T)$values >= 0)
    addError(keyword, "The correlation matrix has negative eigenvalues", condition = indicator)
    colnames(A) <- rownames(A)
  }
  return(A)
}


# Transform into a named vector
transformNamedVector <- function(A, columnValue, columnLabel){
  if(length(columnValue) == 0){
    return(A)
  }
  A <- stats::setNames(object = A[[columnValue]], nm = A[[columnLabel]])
  return(A)
}


# Cast the column into the correct type
transformColumnTypes <- function(Table, colTypes, keyword){
  Table <- copy(Table)
  
  lapplyCatch(colTypes, function(colType){
    set(Table, j = colType$name, value = transformType(Table, column = colType$name, type = colType$type, keyword = keyword))
  })
  return(Table)
}


# Transform the value into a specified type
transformType <- function(Table, column, type, keyword){
  values <- Table[[column]]
  oldValues <- Table[[column]]
  rows <- Table[["rowNumber"]]
  
  # Convert to text
  if(type == "Text"){
    # Empty cells are considered as an empty string.
    values[is.na(values)] <- ""
    return(values)
  }
  
  # Convert to logical
  if(type == "Logical"){
    # Empty or NA are interepreted as FALSE
    values[is.na(values)] <- "No"
    values[values == ""] <- "No"
    index <- !values %in% c("Yes", "No")
    
    addErrorCell(keyword, "The value should be either 'Yes' or 'No'", rows = rows[index], columns = column)
    
    values <- values == "Yes"
    return(values)
  }
  
  # Convert to numeric
  if(type == "Numeric"){
    values <- suppressWarnings(as.numeric(values))
  }else if(type == "Integer"){
    numericValue <- suppressWarnings(as.numeric(values))
    
    # Identify entries which were not integers
    index <-  is.na(numericValue) | (numericValue != as.integer(numericValue))
    values[index] <- NA
    
    values <- as.integer(values)
  }else{
    stop("Incorrect type")
  }
  
  # When the value is NA and was not NA before, then there is information loss while transforming the cell.
  index <- !is.na(oldValues) & is.na(values)
  addErrorCell(keyword, paste0("Can't convert '", oldValues[index], "' into ", ifelse(type == "Numeric", "a real number", "an integer")), rows = rows[index], columns = column)
  return(values)
}
