# Get the submodel
getSubModel <- function(Objects, numberLoad){
  # Check that the risk factors are defined in a consistent way
  checkConsistencyRF(correlationMarket = getValue(Objects, "correlationMarket"),
                     marketVolatility = getValue(Objects, "marketVolatility"),
                     scenarioMacroEconomic = getValue(Objects, "scenarioMacroEconomic"))
  
  # Get target currency
  targetCurrency <- getTargetCurrency(marketRiskFactorsCurrency = getValue(Objects, "marketRiskFactorsCurrency"))
  
  marketCovariance <- getMarketCovariance(correlation = getValue(Objects, "correlationMarket"),
                                          volatility = getValue(Objects, "marketVolatility")
  )
  
  # Maturities are associated with the interest instruments
  RateWithTime <- getRateWithTime(marketRiskFactorsRate = getValue(Objects, "marketRiskFactorsRate"),
                                  maturitiesInterestRate = getValue(Objects, "maturitiesInterestRate"))
  
  # Mapping used for the currency change when the target currency is not CHF 
  mappingDeltaRF <- getCurrencyChangeMapping(Currency = getValue(Objects, "marketRiskFactorsCurrency"),
                                             Delta = getValue(Objects, "marketRiskFactorsDelta"),
                                             refCurrency = getValue(Objects, "referenceCurrency"),
                                             riskFactors = rownames(getValue(Objects, "correlationMarket")),
                                             targetCurrency = targetCurrency
  )
  
  # All instruments, before any adjustment
  rawInstruments <- getRawInstruments(instrumentAsset = getValue(Objects, "instrumentAsset"),
                                      instrumentAssetForward = getValue(Objects, "instrumentAssetForward"),
                                      instrumentFX_Forward = getValue(Objects, "instrumentFX_Forward"),
                                      instrumentFixedIncome = getValue(Objects, "instrumentFixedIncome"),
                                      instrumentLiability = getValue(Objects, "instrumentLiability"),
                                      instrumentDelta = getValue(Objects, "instrumentDelta"),
                                      referenceCurrency = getValue(Objects, "referenceCurrency"),
                                      mappingDeltaRF = mappingDeltaRF
  )
  
  # Check instruments
  checkInstruments(marketRiskFactorsAsset = getValue(Objects, "marketRiskFactorsAsset"),
                   marketRiskFactorsSpread = getValue(Objects, "marketRiskFactorsSpread"),
                   marketRiskFactorsCurrency = getValue(Objects, "marketRiskFactorsCurrency"),
                   rawInstruments = rawInstruments,
                   targetCurrency = targetCurrency,
                   RateWithTime = RateWithTime
  )
  
  # Apply fx change and discounting
  instruments <- getInstruments(refCurrency = getValue(Objects, "referenceCurrency"),
                                initialRate = getValue(Objects, "initialInterestRate"),
                                initialFX = getValue(Objects, "initialFX"),
                                rawInstruments = rawInstruments
  )
  
  # Associate each instrument to a risk factor
  instrumentsMapping <- getInstrumentsMapping(marketRiskFactorsAsset = getValue(Objects, "marketRiskFactorsAsset"),
                                              marketRiskFactorsSpread = getValue(Objects, "marketRiskFactorsSpread"),
                                              marketRiskFactorsCurrency = getValue(Objects, "marketRiskFactorsCurrency"),
                                              marketRiskFactorsDelta = getValue(Objects, "marketRiskFactorsDelta"),
                                              referenceCurrency = getValue(Objects, "referenceCurrency"),
                                              instruments = instruments,
                                              targetCurrency = targetCurrency,
                                              RateWithTime = RateWithTime
  )
  
  # A list of instruments is associated to each standalone category. This is used for the computation of standalone risk.
  instrumentsStandalones <- getInstrumentsStandalones(marketRiskFactorsStandalones = getValue(Objects, "marketRiskFactorsStandalones"),
                                                      instruments = instruments,
                                                      instrumentsMapping = instrumentsMapping,
                                                      marketCovariance = marketCovariance
  )
  
  # Switch for the new credit risk model
  useNewCreditRiskModel <- getValue(Objects, "useNewCreditRiskModel")
  
  # Total progress allocated to the SST-Template (there may be mutliple templates for the participation model) and the variant (there may be multiple variants).
  totalProgressStep <- 1 / numberLoad
  
  # Allocate the total progress for the simualtions between market risk and credit risk
  if(useNewCreditRiskModel){
    progressMarket <- totalProgressStep / 2
    progressCredit <- totalProgressStep / 2
  }else{
    progressMarket <- totalProgressStep
    progressCredit <- 0
  }
  
  lifeModule <- getLifeRisks(correlation = getValue(Objects, "correlationLife"),
                             exposureLife = getValue(Objects, "exposureLife")
  )
  
  nonlifeModule <- getNonLifeRisks(nonlifeType = getValue(Objects, "nonlifeType"),
                                   nonlifeSimulation = getValue(Objects, "nonlifeSimulation"),
                                   nonlifeMu = getValue(Objects, "nonlifeMu"),
                                   nonlifeSigma = getValue(Objects, "nonlifeSigma"),
                                   nonlifeCDF = getValue(Objects, "nonlifeCDF"),
                                   nonLifeSeedAddon = getValue(Objects, "nonLifeSeedAddon"),
                                   captiveCY_deterministic = getValue(Objects, "captiveCY_deterministic"),
                                   captiveCY_stochastic = getValue(Objects, "captiveCY_stochastic"),
                                   captivePY = getValue(Objects, "captivePY")
  )
  
  healthModule <- getHealthRisks(correlation = getValue(Objects, "correlationHealth"),
                                 sensitivityHealth = getValue(Objects, "sensitivityHealth"),
                                 scenarioHealth = getValue(Objects, "scenarioHealth")
  )
  
  
  marketInstrumentsModule <- getMarketInstrumentsModule(instrumentsStandalones, marketCovariance, progressMarket)
  
  marketParticipationModule <- getMarketParticipationModule(participationValue = getValue(Objects, "participationValue"),
                                                            participationVolatility = getValue(Objects, "param_participationVolatility")
  )
  
  creditBaselModule <- getCreditBaselRisks(creditRiskExposure = getValue(Objects, "creditRiskExposure"),
                                           creditRiskExposureReinsurance = getValue(Objects, "creditRiskExposureReinsurance"),
                                           creditRiskExposureHypo = getValue(Objects, "creditRiskExposureHypo"),
                                           param_creditRiskFactor = getValue(Objects, "param_creditRiskFactor"),
                                           useNewCreditRiskModel = getValue(Objects, "useNewCreditRiskModel")
  )
  
  
  creditMertonModule <- getCreditMertonRisks(spreadCurve = getValue(Objects, "spreadCurve"),
                                             LGDMap = getValue(Objects, "mappingLGD"),
                                             migrationMatrix = getValue(Objects, "migrationMatrix"),
                                             initialInterestRate = getValue(Objects, "initialInterestRate1"),
                                             portfolio = getValue(Objects, "portfolioCreditRisk"),
                                             rho = getValue(Objects, "rho"),
                                             initialFX = getValue(Objects, "initialFX"),
                                             refCurrency = getValue(Objects, "referenceCurrency"),
                                             progress = progressCredit
  )
  
  scenarioModule <- getScenarioRisks(participationValue = getValue(Objects, "participationValue"),
                                     scenarioInput = getValue(Objects, "scenarioSST"),
                                     scenarioMacroEconomic = getValue(Objects, "scenarioMacroEconomic"),
                                     instrumentsStandalones = instrumentsStandalones)
  
  # MVM before nhmr effect
  Objects$MVMbeforeNhmr <- getMVM(MVM_CashFlowsLife = getValue(Objects, "MVM_CashFlowsLife"),
                                  MVM_Health = getValue(Objects, "MVM_Health"),
                                  MVM_NonLife = getValue(Objects, "MVM_NonLife"),
                                  referenceCurrency = getValue(Objects, "referenceCurrency"),
                                  initialInterestRate = getValue(Objects, "initialInterestRate"),
                                  volatility = getValue(lifeModule, "volatility"),
                                  correlationLife = getValue(Objects, "correlationLife"),
                                  param_costOfCapital = getValue(Objects, "param_costOfCapital")
  )
  
  # Exchange rate to CHF
  Objects$FX_to_CHF <- getFX_to_CHF(initialFX = getValue(Objects, "initialFX"),
                                    referenceCurrency = getValue(Objects, "referenceCurrency")
  )
  
  # Threshold used to floor the deltaRBC from the daughter company
  Objects$flooringDeltaRBC <- ifelse(getValue(Objects, "daughterApplyLLPO"), -getValue(Objects, "daughterValue"), -Inf)
  
  # Risk as participation indicator
  Objects$riskAsParticipation <- getValue(Objects, "riskAsParticipation")
  
  # All constants
  constantModule <- getConstantModule(expectedFinancialResultTable = getValue(Objects, "expectedFinancialResultTable"),
                                      expectedFinancialResultFactor = getValue(Objects, "expectedFinancialResultFactor"),
                                      expectedInsuranceResult = getValue(Objects, "expectedInsuranceResult"),
                                      RTK_goingConcern = getValue(Objects, "RTK_goingConcern"),
                                      RTK_runOff = getValue(Objects, "RTK_runOff"),
                                      additionalEffectsOnTC = getValue(Objects, "additionalEffectsOnTC"),
                                      daughterValue = getValue(Objects, "daughterValue")
  )
  
  # Modules to be simulated
  modules <- list(life = lifeModule,
                  nonlife = nonlifeModule,
                  health = healthModule,
                  marketInstruments = marketInstrumentsModule,
                  marketParticipation = marketParticipationModule,
                  creditBasel = creditBaselModule,
                  creditMerton = creditMertonModule,
                  scenario = scenarioModule,
                  constant = constantModule
  )
  
  
  # Dependency structure
  dependencyRisks <- getDependencyRisks(correlation = getValue(Objects, "correlationRiskAggregation"))
  dependencyCredit <- getDependencyCredit(corr = getValue(Objects, "corr"))
  dependencyMarket <- getDependencyMarket()
  
  # Dependency structure to be simulated 
  dependency <- list(risks = dependencyRisks,
                     credit = dependencyCredit,
                     market = dependencyMarket)
  
  modules <- namedObject(modules, "Module from the SST-Model", "Module")
  dependency <- namedObject(dependency, "Dependency structure from the SST-Model", "Dependency")
  
  Model <- list(Modules = modules,
                Dependency = dependency,
                Objects = Objects
  )
  
  retClass(Model, "sstModel")
}


# Get the model
getModel <- function(path, variant, numberLoad){
  # Iterate over all SST-Templates to load the objects
  objectsAllTemplates <- lapplyCatch(path, getObjects, variant = variant, numberLoad = numberLoad)
  
  # Take into account the interdependency between the mother and the daughter
  processedObjectsAllTemplates <- getProcessedObjectsAllTemplates(objectsAllTemplates)
  
  # Generate the submodel for mother (and daughters if any)
  model <- lapply(processedObjectsAllTemplates, getSubModel, numberLoad = numberLoad)
  
  # Adjustment for the credit risk
  allCounterparties <- getAllCounterparties(model)
  for(i in seq_along(model)){
    model[[i]]$Modules$creditMerton$counterparties <- allCounterparties
  }
  
  retClass(model, "sstCalculation")
}

