# Non-life risks class
getNonLifeRisks <- function(nonlifeType, nonlifeSimulation, nonlifeMu, nonlifeSigma, nonlifeCDF, nonLifeSeedAddon, captiveCY_deterministic, captiveCY_stochastic, captivePY){
  NL_none <- getParam("NL_none")
  NL_sim <- getParam("NL_sim")
  NL_log <- getParam("NL_log")
  NL_cdf <- getParam("NL_cdf")
  NL_capt <- getParam("NL_capt")
  
  indic <- list(nonlifesimulation = nrow(nonlifeSimulation) > 0,
                nonlifemu = !is.na(nonlifeMu),
                nonlifesigma = !is.na(nonlifeSigma),
                nonlifecdf = nrow(nonlifeCDF) > 0,
                nonlifecaptiveCY = nrow(captiveCY_deterministic) + nrow(captiveCY_stochastic) > 0,
                nonlifecaptivePY = nrow(captivePY) > 0
  )
  
  message <- paste0("The input will be ignored as a different non-life type is selected (type selected: '", nonlifeType, "')")
  addError("nonlifeSimulation", message, condition = (nonlifeType != NL_sim) && indic$nonlifesimulation, type = "Warning")
  addError("nonlifeMu", message, condition = (nonlifeType != NL_log) && indic$nonlifemu, type = "Warning")
  addError("nonlifeSigma", message, condition = (nonlifeType != NL_log) && indic$nonlifesigma, type = "Warning")
  addError("nonlifeCDF", message, condition = (nonlifeType != NL_cdf) && indic$nonlifecdf, type = "Warning")
  addError("captiveCY", message, condition = (nonlifeType != NL_capt) && indic$nonlifecaptiveCY, type = "Warning")
  addError("captivePY", message, condition = (nonlifeType != NL_capt) && indic$nonlifecaptivePY, type = "Warning")
  
  message <- paste0("The input to simulate the non-life risk is missing. Please ensure that the input for the risk type '", nonlifeType, "' is provided or set the risk type to '", NL_none, "' if no non-life risk should be simulated")
  addError("nonlifeSimulation", message, condition = (nonlifeType == NL_sim) && !indic$nonlifesimulation)
  addError("nonlifeMu", message, condition = (nonlifeType == NL_log) && !indic$nonlifemu)
  addError("nonlifeSigma", message, condition = (nonlifeType == NL_log) && !indic$nonlifesigma)
  addError("nonlifeCDF", message, condition = (nonlifeType == NL_cdf) && !indic$nonlifecdf)
  addError("captiveCY", message, condition = (nonlifeType == NL_capt) && !indic$nonlifecaptiveCY && !indic$nonlifecaptivePY)
  
  if(nonlifeType == NL_sim){
    # Simulation vector
    param <- list(simulation = nonlifeSimulation$simulation)
    param <- retClass(param, "NL_sim")
    
  }else if(nonlifeType == NL_log){
    # Lognormal parameters
    param <- list(mu = nonlifeMu, sigma = nonlifeSigma)
    param <- retClass(param, "NL_log")
    
  }else if(nonlifeType == NL_cdf){
    # Distribution function
    param <- list(cdf = nonlifeCDF)
    param <- retClass(param, "NL_cdf")
    
  }else if(nonlifeType == NL_none){
    # No risk
    param <- list()
    param <- retClass(param, "NL_none")
    
  }else if(nonlifeType == NL_capt){
    # Make a copy since they will be edited by reference
    captiveCY_stochastic <- copy(captiveCY_stochastic)
    captivePY <- copy(captivePY)
    
    # Normal losses
    row <- captiveCY_stochastic[allConsistent(normalLossNumber, normalLossSize, normalLossSigma) == FALSE, rowNumber]
    addErrorCell("captiveCY_stochastic", "Normal loss parameters should be either all provided or all left empty.", columns = "normalLossNumber", rows = row)
    
    # Large losses
    row <- captiveCY_stochastic[allConsistent(largeLossNumber, largeLossThreshold, largeLossShape) == FALSE, rowNumber]
    addErrorCell("captiveCY_stochastic", "Large loss parameters should be either all provided or all left empty.", columns = "normalLossNumber", rows = row)
    
    # No losses
    row <- captiveCY_stochastic[is.na(normalLossNumber) & is.na(largeLossNumber), rowNumber]
    addErrorCell("captiveCY_stochastic", "Please provide at least the normal loss parameters or the large loss parameters.", columns = "normalLossNumber", rows = row)
    
    # Maximum possible loss
    row <- captiveCY_deterministic[maximumPossibleLoss < expectedLoss, rowNumber]
    addErrorCell("captiveCY_deterministic", "The maximum possible loss should be larger than the expected loss", columns = "maximumPossibleLoss", rows = row)
    
    captiveCY_stochastic[, ':=' (normalLossScale = normalLossSigma^2/normalLossSize, normalLossShape = normalLossSize^2/normalLossSigma^2)]
    captiveCY_stochastic[is.na(AAL), AAL := Inf]
    captiveCY_stochastic[is.na(EEL), EEL := Inf]
    captiveCY_stochastic[is.na(QS), QS := 1]
    
    captivePY[, sd := sqrt(log(1+cov^2))]
    captivePY[, mu := log(reserve)-0.5*sd^2]
    captivePY[, ':=' (reserve = NULL, cov = NULL)]
    
    CY <- list(stochastic = captiveCY_stochastic, deterministic = captiveCY_deterministic[, sum(maximumPossibleLoss - expectedLoss)])
    CY <- retClass(CY, "NL_captCY")
    
    PY <- list(stochastic = captivePY)
    PY <- retClass(PY, "NL_captPY")
    
    param <- list(CY = CY, PY = PY)
    param <- retClass(param, "NL_capt")
  }
  
  x <- list(param = param, nonLifeSeedAddon = nonLifeSeedAddon)
  retClass(x, "nonlifeModule")
}
