context("market")



test_that("calculateStandalone", {
  
  # Zero exposure
  instrument <- data.table(standaloneTypes = "all",
                           instrumentId = 3,
                           instrument = "Delta",
                           exposure = 0,
                           exponentialValuation = FALSE,
                           informationJoinedRF = "asset",
                           factorId = "RF4",
                           name = "MSCI JP",
                           scale = 2,
                           constant = -5,
                           progress = 1
  )
  sim <- calculateStandalone("all", data.table(`MSCI JP`=2), instrument)
  expect_identical(sim, 0)
  
  
  ## Delta case
  # positive RF
  instrument <- data.table(standaloneTypes = "all",
                           instrumentId = 3,
                           instrument = "Delta",
                           exposure = 5,
                           exponentialValuation = FALSE,
                           informationJoinedRF = "asset",
                           factorId = "RF4",
                           name = "MSCI JP",
                           scale = 2,
                           constant = -5, 
                           progress = 1
  )
  sim <- calculateStandalone("all", data.table(`MSCI JP`=2), instrument)
  expect_identical(sim, 20)
  # negative RF
  sim <- calculateStandalone("all", data.table(`MSCI JP`=-1), instrument)
  expect_identical(sim, -10)
  
  ## 1 risk factor
  instrument <- data.table(standaloneTypes = "all",
                           instrumentId = 1,
                           instrument = "Equity",
                           exposure = 1,
                           exponentialValuation = TRUE,
                           informationJoinedRF = "asset",
                           factorId = "RF4",
                           name = "MSCI JP",
                           scale = 2,
                           constant = 2,
                           progress = 1
  )
  sim <- calculateStandalone("all", data.table(`MSCI JP`=-1), instrument)
  expect_identical(sim, 0)
  
  
  ## 2 risk factors
  # null scale
  instrument <- data.table(standaloneTypes = "all",
                           instrumentId = 1,
                           instrument = "Equity",
                           exposure = 1,
                           exponentialValuation = TRUE,
                           informationJoinedRF = c("asset", "currency"),
                           factorId = c("RF4", "RF21"),
                           name = c("MSCI JP", "JPYCHF"),
                           scale = 0,
                           constant = 0,
                           progress = 0.5
  )
  sim <- calculateStandalone("all", data.table(`MSCI JP`=-1, `JPYCHF`=1), instrument)
  expect_identical(sim, 0)
  # other case
  instrument <- data.table(standaloneTypes = "all",
                           instrumentId = 1,
                           instrument = rep("Equity", 2),
                           exposure = 1,
                           exponentialValuation = TRUE,
                           informationJoinedRF = c("asset", "currency"),
                           factorId = c("RF4", "RF21"),
                           name = c("MSCI JP", "JPYCHF"),
                           scale = 1,
                           constant = 1,
                           progress = 0.5
  )
  sim <- calculateStandalone("all", data.table(`MSCI JP`=-1, `JPYCHF`=1), instrument)
  expect_identical(sim, exp(1)-1)
  
  
  ## 3 risk factors - different scale
  instrument <- data.table(standaloneTypes = "all",
                           instrumentId = 1,
                           instrument = "Equity",
                           exposure = 1,
                           exponentialValuation = TRUE,
                           informationJoinedRF = c("asset", "currency", "currency"),
                           factorId = c("RF4", "RF21", "RF18"),
                           name = c("MSCI JP", "JPYCHF", "EURCHF"),
                           scale = 1:3,
                           constant = 1,
                           progress = 1/3
  )
  sim <- calculateStandalone("all", data.table(`MSCI JP`=-1, `JPYCHF`=1/2, `EURCHF`=-1/3), instrument)
  expect_identical(sim, 0)
  
  ## 4 risk factors
  # 1 simulation per risk factor
  instrument <- data.table(standaloneTypes = "all",
                           instrumentId = 1,
                           instrument = "Fixed Income",
                           exposure = 1,
                           exponentialValuation = TRUE,
                           informationJoinedRF = c( "spread", "currency", "rate", "currency"),
                           factorId = c("RF48", "RF18", "RF25", "RF21"),
                           name = c("EUR Spread AA", "EURCHF", "EUR 2Y", "JPYCHF"),
                           scale = 1,
                           constant = 0,
                           progress = 0.25
  )
  sim <- calculateStandalone("all", data.table(`EUR Spread AA`=1, `JPYCHF`=1, `EURCHF`=1, `EUR 2Y`=1), instrument)
  expect_identical(sim, exp(4)-1)
  
  
  # case with 2 simulations per risk factor
  RF <- data.table(`EUR 2Y` = c(0,1),
                   `EUR Spread AA` = c(0,-1),
                   `EURCHF` = c(0,1),
                   `JPYCHF` = c(0, -2),
                   `MSCI JP` = c(0,100)
  )
  sim <- calculateStandalone("all", RF, instrument)
  expect_identical(sim, c(0, exp(-1)-1))
})



test_that("calculateInitialSpread", {
  # Exact match
  expect_identical(calculateInitialSpread(marketValue = 10, times = 1:10, coupons = rep(1, 10), riskFree = rep(0, 10)), 0)
  
  # NaN produced by out of range value
  expect_true(is.nan(calculateInitialSpread(marketValue = 10, times = 1:10, coupons = rep(1/1000, 10), riskFree = rep(0, 10))))
  
  # Correct spread
  expect_equal(calculateInitialSpread(marketValue = sum(exp(-0.01 * (1:10))), times = 1:10, coupons = rep(1, 10), riskFree = rep(0, 10)), 0.01)
  
  # Correct extreme spread (positive)
  expect_equal(calculateInitialSpread(marketValue = sum(exp(-10 * (1:10))), times = 1:10, coupons = rep(1, 10), riskFree = rep(0, 10)), 10)
  # Correct extreme spread (negative)
  expect_equal(calculateInitialSpread(marketValue = sum(exp(0.30 * (1:10))), times = 1:10, coupons = rep(1, 10), riskFree = rep(0, 10)), -0.3)
  
  # non-null riskFree
  expect_equal(calculateInitialSpread(marketValue = 13, times = 1:4, coupons = c(rep(1, 3),10), riskFree = rep(0.1, 4)), -0.1, tolerance=1e-10)
})




# test_that("simulate.MarketParticipation", {
#   exposure <- 100
#   sd <- 5
#   object <- list(exposure=exposure, sd=sd)
#   nsim <- 5
#   seed <- 1
#   setSeed(seed, "Participation")
#   expect_identical(exposure * expm1(stats::rnorm(n = nsim, sd = sd) - 0.5 * (sd^2)), 
#                    simulate.MarketParticipation(object, nsim, seed))
# })



# test_that("getMarketParticipation", {
#   partValue <- 10
#   partVola <- 1
#   x <- list(exposure=partValue, sd=partVola)
#   expect_identical(retClass(x, "MarketParticipation"), getMarketParticipation(partValue, partVola))
# })


# test_that("getMarketInstruments", {
#   instrument <- data.table(standaloneTypes = "all",
#                            instrumentId = 3,
#                            instrument = "Delta",
#                            exposure = 0,
#                            exponentialValuation = FALSE,
#                            informationJoinedRF = "asset",
#                            factorId = "RF4",
#                            name = "MSCI JP",
#                            scale = 2,
#                            constant = 0,
#                            progress = 1
#   )
#   standalonesList <- stats::setNames(nm = c("all", getParam("standaloneTypes")))
#   Sigma <- matrix(c(1,0,0,1), ncol=2)
#   mu <- c(0,0)
#   x <- list(instruments = instrument, standalonesList = standalonesList, Sigma = Sigma, mu = mu)
#   expect_identical(retClass(x, "MarketInstruments"), getMarketInstruments(instrument, Sigma))
# })



# test_that("getMarketRisks", {
#   marketCovariance <- matrix(c(1,0,0,1), ncol=2)
#   participationValue <- 10
#   participationVolatility <- 1
#   instrumentsStandalones <- 3
#   x <- list(instruments = getMarketInstruments(instrumentsStandalones, marketCovariance),
#             participation = getMarketParticipation(participationValue, participationVolatility))
#   expect_identical(retClass(x, "marketModule"), getMarketRisks(instrumentsStandalones, marketCovariance, 
#                                                                participationValue, participationVolatility))
# })



test_that("calculateStandalones", {
  standaloneType <- "all"
  instruments <- data.table(instrumentId = 1,
                            instrument = "Equity",
                            exposure = 1,
                            exponentialValuation = TRUE
  )
  instrumentsMapping <- data.table(informationJoinedRF = c("asset","currency"),
                                   instrumentId = 1,
                                   factorId = c("RF4", "RF21"),
                                   name = c("MSCI JP", "JPYCHF"),
                                   scale = 1
  )
  MR_FactorsStandalones <- data.table(factorId = c("RF4", "RF21"),
                                      rowNumber = c(7,8),
                                      standalones = c("equity", "currency")
  )
  marketCovariance <- matrix(c(1,0,0,1), ncol=2, dimnames = list(c("MSCI JP", "JPYCHF"),c("MSCI JP", "JPYCHF")))
  x <- copy(instruments[instrumentsMapping, on = "instrumentId", nomatch = 0])
  x[, constant:= c(-1,-1)]
  setindex(x, "exponentialValuation")
  
  y <- calculateStandalones(standaloneType, instruments, instrumentsMapping, MR_FactorsStandalones, marketCovariance)
  expect_identical(x,y)
})



# test_that("simulate.MarketInstruments", {
#   Sigma <- matrix(c(1,0,0,1), ncol=2)
#   mu <- c(0,0)
#   instrument <- data.table(standaloneTypes = "all",
#                            instrumentId = 3,
#                            instrument = "Delta",
#                            exposure = 0,
#                            exponentialValuation = FALSE,
#                            informationJoinedRF = "asset",
#                            factorId = "RF4",
#                            name = "MSCI JP",
#                            scale = 2,
#                            constant = -5,
#                            progress = 1
#   )
#   st_list <- list("all"="all")
#   object <- list(mu=mu, Sigma=Sigma, standalonesList=st_list, instruments=instrument)
#   nsim <- 10
#   seed <- 1
#   set.seed(seed)
#   RF <- data.table(MASS::mvrnorm(n = nsim, mu = mu, Sigma = Sigma))
#   names(RF) <- c("MSCI JP", "FOO")
#   sim <- as.data.table(lapply(st_list, calculateStandalone, RF = RF, instruments = instruments, centered = TRUE, participationValue = 0))
#   
#   result <- list(sim = sim[], RF = RF[])
#   simulate.MarketInstruments(object, nsim, seed)
# })



